from ._infer_statistics import *
from .EventComposite import *
from ._find_events import *
from ._process_events import *
from ._manage import *
from ._utils import *
from ._infer_metadata import *
from ._extreme_prob import *
from ._ssw_compendium import *



def stack_fc(d, reset_index=True):
    if reset_index:
        return d.stack(fc=("reftime", "hc_year", "number")).reset_index("fc")
    else:
        return d.stack(fc=("reftime", "hc_year", "number"))


def stack_ensfc(d, reset_index=True):
    if reset_index:
        return d.stack(fc=("reftime", "hc_year")).reset_index("fc")
    else:
        return d.stack(fc=("reftime", "hc_year"))



from ecmwfapi import ECMWFDataServer


class S2SDownloader:
    DEFAULT_REQUEST = {
        "class": "s2",
        "dataset": "s2s",
        "expver": "prod",
        # "date": '2021-02-04',
        "expver": "prod",
        # "levelist": plevs,
        "levtype": "pl",
        "model": "glob",
        # "origin": "ecmf",
        # "step": "0/24",
        # "stream": "enfo",
        "time": "00:00:00",
        # "area": ["65", "-10", "50", "130"],
        "type": "cf",
        "format": "netcdf",
        # "grid": "2.5/2.5",
        # "target": target,
    }

    request = {}

    def __init__(self):
        pass

    def retrieve(self, request):
        print("Target: ", request["target"])

        if True:
            server = ECMWFDataServer()

            server.retrieve(
                request
            )



import pandas as pd

from s2stools import utils
import numpy as np


class Rt:
    request_rt = {}

    def __init__(self):
        self.request_rt["stream"] = "enfo"


class RtCf(Rt):
    request = {}

    def __init__(self, rtdate=None):
        super().__init__()
        self.request = super().request_rt
        self.request["type"] = "cf"


class RtPf(Rt):
    request = {""}

    def __init__(self, rtdate=None):
        super().__init__()
        self.request = super().request_rt
        self.request["type"] = "pf"
        self.request["number"] = utils.list_to_string(self.number_all())

    def number_all(self):
        """
        Return all available ensemble members for realtime perturbed forecasts.
        Returns: String numbering the Members.

        """
        return range(1, 51)


class Hc:
    request_hc = {}

    def __init__(self, rtdate):
        self.request_hc["hdate"] = utils.list_to_string(self.hdate_all(rtdate))
        self.request_hc["stream"] = "enfh"

    def hdate_all(self, date):
        """
        All available hindcast dates.
        Args:
            date (np.datetime64): Realtime date to which the hindcast refers.

        Returns: np.ndarray(datetime64[D]) of all hindcast initialization dates

        """
        end = pd.to_datetime(date)
        if (end.month == 2) & (end.day == 29):
            start = end - pd.DateOffset(years=20) - pd.DateOffset(days=1)
        else:
            start = end - pd.DateOffset(years=20)

        hdates = pd.date_range(start, periods=20, freq=pd.DateOffset(years=1))
        return np.array(hdates, dtype="datetime64[D]")


class HcCf(Hc):
    request = {}

    def __init__(self, rtdate):
        super().__init__(rtdate=rtdate)
        self.request = super().request_hc
        self.request["type"] = "cf"
        self.request["stream"] = "enfh"
        # self.request["alias"] = "hccf"


class HcPf(Hc):
    request = {}

    def __init__(self, rtdate):
        super().__init__(rtdate=rtdate)
        self.request = super().request_hc
        self.request["type"] = "pf"
        self.request["stream"] = "enfh"
        # self.request["alias"] = "hcpf"
        self.request["number"] = self.number_all()

    def number_all(self):
        """
        Return all available ensemble members for hindcast perturbed forecasts.
        Returns: String numbering the Members.

        """
        return utils.list_to_string(range(1, 11))



def ssw_compendium_event_dates(column='ERA-Interim'):
    # read html table from SSW compendium
    result = pd.read_html(
        "https://csl.noaa.gov/groups/csl8/sswcompendium/majorevents.html"
    )
    df_raw = result[0]

    # parse dates
    def parse_dates(x):
        if "-" in str(x):
            x = pd.to_datetime(str(x), format="%d-%b-%y")
            x = x.replace(year=x.year - 100) if x.year > datetime.now().year else x
            return x
        else:
            return x

    df = df_raw.applymap(parse_dates)

    # filter for valid dates
    erai_dates = df[column][
        df[column].apply(lambda x: isinstance(x, pd.Timestamp))
    ]

    return erai_dates



from unittest import TestCase
import numpy as np
import pandas as pd

from s2stools.download.ecmwf import S2SDownloaderECMWF
from s2stools.download.ecmwf import model_setup


class TestS2SDownloaderECMWF(TestCase):

    def test_retreive(self):
        if False:  # set to true to test download via ECMWF API
            dl = S2SDownloaderECMWF()
            dates = np.arange("2020-11-15", "2020-12-20", dtype="datetime64[D]")

            dl.retreive(
                param=["u", "v"],
                file_descr="uv",
                reftime=dates[:5],
                plevs=[1000],
                step=[0, 24],
                path="../../data",
                grid="2.5/2.5",
                rt_cf_kwargs={},
                rt_pf_kwargs=dict(number="1/2/3"),
                # hc_cf_kwargs=dict(hdate="2000-11-16/2001-11-16"),
                hc_pf_kwargs=dict(number="1/2"),
                write_info_file=True
            )

    def test_retreive_rt_cf_only(self):
        if False:  # set to true to test download via ECMWF API
            dl = S2SDownloaderECMWF()
            dates = np.arange("2020-11-15", "2020-12-20", dtype="datetime64[D]")

            dl.retreive(
                param=["u", "v"],
                file_descr="uv",
                reftime=dates[:5],
                plevs=[1000],
                step=[0, 24],
                path="../../data",
                grid="2.5/2.5",
                rt_cf_kwargs={},
                rt_pf_kwargs=dict(number="1/2/3", skip=True),
                hc_cf_kwargs=dict(skip=True), #hdate="2000-11-16/2001-11-16"
                hc_pf_kwargs=dict(number="1/2", skip=True),
                write_info_file=True
            )


    def test_filter_reftimes(self):
        dl = S2SDownloaderECMWF()
        dates = np.arange("2021-11-15", "2021-12-15", dtype="datetime64[D]")

        valid_dates = dl.filter_reftimes(dates)

        self.assertEqual(len(valid_dates), 9)
        self.assertIn(np.datetime64("2021-11-15"), valid_dates)
        self.assertNotIn(np.datetime64("2021-11-16"), valid_dates)

    def test_hdates_all(self):
        leap_yr_jan = np.datetime64("2016-01-16")
        leap_yr_nov = np.datetime64("2016-11-16")
        non_leap_yr_jan = np.datetime64("2017-01-16")
        non_leap_yr_nov = np.datetime64("2017-11-16")
        leap_yr_29feb = np.datetime64("2016-02-29")

        test_dates = [leap_yr_jan, leap_yr_nov, non_leap_yr_jan, non_leap_yr_nov, leap_yr_29feb]
        for d in test_dates:
            d_pd = pd.to_datetime((d))
            client = model_setup.Hc(d)
            hdates = client.hdate_all(d)
            # days and months of hindcast dates are equal
            self.assertEqual(0, np.max(np.abs(np.diff(pd.to_datetime(hdates).day))), msg="1")
            self.assertEqual(0, np.max(np.abs(np.diff(pd.to_datetime(hdates).month))), msg="2")
            # months and days are equal to realtime date (except: 29 feb)
            if (d_pd.month == 2) & (d_pd.day == 29):
                self.assertEqual(pd.to_datetime(hdates[0]).month, 2)
                self.assertEqual(pd.to_datetime(hdates[0]).day, 28)
            else:
                self.assertEqual(pd.to_datetime(d).day, pd.to_datetime(hdates[0]).day, msg="3")
                self.assertEqual(pd.to_datetime(d).month, pd.to_datetime(hdates[0]).month, msg="4")



from unittest import TestCase
from s2stools import events, process, plot

import matplotlib.pyplot as plt


class TestEvents(TestCase):

    def test_find_ssw(self):
        ds = process.open_files(path_pattern="../../data/s2s_ecmwf_uv_20*")
        u60_10hPa_proxy = ds.u.sel(latitude=60).mean("longitude")
        # u_fc = process.stack_fc(u60_10hPa_proxy)
        event_list = events.find_ssw(u60_10hPa_proxy, buffer_start=0, buffer_end=0, require_westwind_start=1)
        comp = events.composite_from_eventlist(event_list, ds)
        u60_10hPa_proxy_comp = comp.u.mean("longitude").sel(latitude=60)
        self.assertGreater(u60_10hPa_proxy_comp.sel(lagtime="-1D").min("i").values, 0)
        self.assertLess(u60_10hPa_proxy_comp.sel(lagtime="0D").max("i").values, 0)

    def test_eventdict_to_json(self):
        ds = process.open_files(path_pattern="../../data/s2s_ecmwf_uv_20*")
        u60_10hPa_proxy = ds.u.sel(latitude=60).mean("longitude")
        u_fc = process.stack_fc(u60_10hPa_proxy)
        eventdict = events.find_ssw(u60_10hPa_proxy, buffer_start=0, buffer_end=0, require_westwind_start=1)
        path = "../data/events/ssw/pseudossw.json"
        events.eventdict_to_json(eventdict, path, split_reftimes=True)

    def test_composite_from_json(self):
        ds = process.open_files(path_pattern="../../data/s2s_ecmwf_uv_20*")
        comp = events.composite_from_json("../../data/events/ssw/*.json", ds)
        self.assertIn("i", list(comp.dims))

    def test_rename_eventlist_key(self):
        eventlist = [{'fc': {'reftime': '2021-01-21', 'hc_year': -20, 'number': 0},
                      'days_since_init': 19},
                     {'fc': {'reftime': '2021-01-21', 'hc_year': -20, 'number': 3},
                      'days_since_init': 30}]
        new_eventlist = events.rename_eventlist_key(eventlist, {"days_since_init": "leadtime"})
        self.assertNotIn("leadtime", eventlist[0].keys())
        self.assertIn("days_since_init", eventlist[0].keys())
        self.assertIn("leadtime", new_eventlist[0].keys())
        self.assertNotIn("days_since_init", new_eventlist[0].keys())

    def test_format_eventlist_dayssinceinit_to_pdtimedeltastamp(self):
        eventlist = [{'fc': {'reftime': '2021-01-21', 'hc_year': -20, 'number': 0},
                      'days_since_init': 19},
                     {'fc': {'reftime': '2021-01-21', 'hc_year': -20, 'number': 3},
                      'days_since_init': 30}]
        new_eventlist = events.format_eventlist_dayssinceinit_to_pdtimedeltastamp(eventlist)
        print(new_eventlist)

    def test_EventComposite(self):
        # load data
        ds = process.open_files(path_pattern="../../data/s2s_ecmwf_uv_20*")
        # SSW index
        u60_10hPa_proxy = ds.u.sel(latitude=60).mean("longitude")  # .sel(hc_year=slice(-5, 0))
        # find events
        event_list = events.find_ssw(u60_10hPa_proxy, buffer_start=0, buffer_end=0, require_westwind_start=1)
        # create EventComposite
        ec = events.EventComposite(data=ds, event_jsons_path=event_list, descr="SSW", model="ecmwf")
        self.assertIsInstance(ec, events.EventComposite)
        # plot
        # plot.composite_overview(ec.comp.sel(latitude=60).mean("longitude"))
        # plt.show()
        # plt.close()
        # ec.plot_eventinfo()
        # plt.show()
        # plt.close()



import xarray as xr
import numpy as np
from s2stools import utils
import glob


def open_files(cf=None, pf=None, chc=None, phc=None, path_pattern=None, max_lt=46):
    """
    TODO: Can we autodetect max_lt?
    TODO: Compare to process_s2s_ukmo – possible to generalize?
    Args:
        cf ():
        pf ():
        chc ():
        phc ():
        path_pattern ():
        max_lt ():

    Returns:

    """
    if path_pattern:
        cf = path_pattern + (cf or "_rt_cf.nc")
        pf = path_pattern + (pf or "_rt_pf.nc")
        chc = path_pattern + (chc or "_hc_cf.nc")
        phc = path_pattern + (phc or "_hc_pf.nc")

    # obtain lead time
    valid_rt = cf or pf
    if valid_rt is None:
        raise NotImplementedError("Without reading realtime forecasts, cannot infer leadtime.")
    print("realtime path: ", valid_rt)
    onefile_path = glob.glob(valid_rt)[0]
    onefile_ds = xr.open_dataset(onefile_path)
    lt = (onefile_ds.time.values - onefile_ds.time.values[0]).astype("timedelta64[h]")
    if (lt.astype("int") % 24).max() == 0:
        # lead time is given in daily resolution
        lt = lt.astype("timedelta64[D]")

    def preprocess(x, lt):
        """
        -> Todo
        Args:
            x ():
            lt (int): lead time

        Returns:

        """
        max_lt = lt.max()  # maximum leadtime
        n_lt = len(lt)  # n leadtime steps
        lt_ts = lt[1] - lt[0]  # leadtime timestep
        hindcast = True if len(x.time) > n_lt else False
        reftime = utils.add_years(x.time[0].values, 20) if hindcast else x.time[0].values
        x = x.assign_coords(reftime=reftime)

        bins = np.append(x.time[::n_lt], x.time[-1].values + lt_ts)
        bin_labels = np.arange(-20, 0) if hindcast else np.array([0])
        groups = x.groupby_bins(
            "time", bins=bins, labels=bin_labels, include_lowest=True, right=False
        )
        hcyear_collection = [g[1].assign_coords(hc_year=int(g[0])) for g in groups]

        res = []
        for hcy in hcyear_collection:
            fc_start = utils.add_years(reftime, int(hcy.hc_year))
            hcy_timeslice = (
                hcy.sel(time=slice(fc_start, fc_start + max_lt))
                    .assign_coords(time=lt)  # np.arange(max_lt+1)
                    .rename(time="leadtime")
            )
            res.append(hcy_timeslice)
        x = xr.concat(res, dim="hc_year")

        return x

    rt = []
    hc = []

    if cf:
        print("cf\t", end=" ")
        s2s_cf = xr.open_mfdataset(
            cf,
            concat_dim="reftime",
            preprocess=lambda x: preprocess(x, lt=lt),
            combine="nested",
        )
        s2s_cf = s2s_cf.assign_coords(number=0).expand_dims("number")
        rt.append(s2s_cf)
        print("")
    if pf:
        print("pf\t", end=" ")
        s2s_pf = xr.open_mfdataset(
            pf,
            concat_dim="reftime",
            preprocess=lambda x: preprocess(x, lt=lt),
            combine="nested",
        )
        rt.append(s2s_pf)
        print("")
    if chc:
        print("chc\t", end=" ")
        s2s_chc = xr.open_mfdataset(
            chc,
            concat_dim="reftime",
            preprocess=lambda x: preprocess(x, lt=lt),
            combine="nested",
        )
        s2s_chc = s2s_chc.assign_coords(number=0).expand_dims("number")
        hc.append(s2s_chc)
        print("")
    if phc:
        print("phc\t", end=" ")
        s2s_phc = xr.open_mfdataset(
            phc,
            concat_dim="reftime",
            preprocess=lambda x: preprocess(x, lt=lt),
            combine="nested",
        )
        hc.append(s2s_phc)
        print("")

    res = []
    if len(rt) != 0:
        rt_merged = xr.concat(rt, dim="number")
        res.append(rt_merged)
    if len(hc) != 0:
        hc_merged = xr.concat(hc, dim="number")
        res.append(hc_merged)
    return xr.concat(res, dim="hc_year").sortby("hc_year")



import json
import pandas as pd
from glob import glob
from pathlib import Path


def eventlist_to_dict(events):
    """
    events should have the shape:
    [
        ((reftime, hc_year, number), leadtime),
        ((reftime, hc_year, number), leadtime),
        ...
    ]
    """
    event_dict = [
        {
            "fc": {
                "reftime": reftime.date().isoformat(),
                "hc_year": int(hc_year),
                "number": int(number),
            },
            # "leadtime": int(leadtime),
            "leadtime": pd.Timedelta(leadtime).isoformat(),
        }
        for (reftime, hc_year, number), leadtime in events
    ]

    return event_dict


def eventdict_to_json(event_dict, path, short_path=False, split_reftimes=False):
    DEFAULT_DIR = (
        "/project/meteo/work/Jonas.Spaeth/Master-Thesis/processed-data/json/s2s_events/"
    )

    def save_dict(e_d, path):
        Path(path.rpartition("/")[0]).mkdir(parents=True, exist_ok=True)
        with open(path, "w") as fout:
            json.dump(e_d, fout)

    if not split_reftimes:
        path = DEFAULT_DIR + path if short_path else path
        save_dict(event_dict, path)
    else:
        event_dict_splitted = events_split_reftime(event_dict)
        for ref in event_dict_splitted:
            ref_str = ref["reftime"]
            events_oneref = ref["events"]
            path_splitted = (
                DEFAULT_DIR
                + path.split(".json")[0]
                + "_ref{}.json".format(ref_str.replace("-", ""))
                if short_path
                else path.split(".json")[0]
                     + "_ref{}.json".format(ref_str.replace("-", ""))
            )
            save_dict(events_oneref, path_splitted)


def events_split_reftime(eventdict):
    """
    convert from
    event_dict = [{"fc": {"reftime": , "hc_year": , "number": }, "leadtime": }, {...}]
    to
    event_dict_splitted = [{"reftime": , "events": [{"fc": {"reftime": , "hc_year": , "number": }, "leadtime": }, {...}]}, {...}]
    """
    event_dict_splitted = []

    for e in eventdict:
        ref = e["fc"]["reftime"]
        idx = next(
            (
                index
                for (index, d) in enumerate(event_dict_splitted)
                if d.get("reftime") == ref
            ),
            None,
        )
        if idx != None:
            # reftime exists
            event_dict_splitted[idx]["events"].append(e)
        else:
            # reftime does not exist yet
            event_dict_splitted.append({"reftime": ref, "events": [e]})

    return event_dict_splitted


def eventlist_from_json(path):
    event_list = []
    for f in glob(path):
        with open(f) as infile:
            event_list = event_list + json.load(infile)
    return event_list



import numpy as np
import datetime
import json
from datetime import datetime
from .. import S2SDownloader
from .model_setup import *

class S2SDownloaderECMWF(S2SDownloader):
    DEFAULT_REQUEST_ECMWF = {
        "origin": "ecmf"
    }
    STEP_ALL = utils.list_to_string(np.arange(0, 1104 + 1, 24))
    MODEL_NAME = "ecmwf"

    request = {}

    def __init__(self):
        S2SDownloader.__init__(self)
        self.request = super().DEFAULT_REQUEST
        self.request.update(self.DEFAULT_REQUEST_ECMWF)

    def retreive(self, param, reftime, plevs, file_descr, path="./", area=None, step="all", exact_reftime=False,
                 write_info_file=True,
                 rt_cf_kwargs={},
                 rt_pf_kwargs={}, hc_cf_kwargs={}, hc_pf_kwargs={}, **kwargs):
        """
        Download S2S ECMWF data.
        Args:
            param (): Parameters to download.
            reftime (): Realtime Dates.
            plevs (): Pressure Levels.
            file_descr (): Description of fields that will show up in the file name.
            path (): Directory to store the output.
            area (str): Area in the format "N/W/S/E"
            step (): Forecast lead times; if "all" then download all available in daily resolution.
            exact_reftime (bool): If true, assert that the given reftime dates are actual model initialization dates, if False, then automatically drop the invalid dates.
            write_info_file (bool): If true, write a .json file into the target directory with info about request.
            rt_cf_kwargs (dict): Additional request keywords passed to realtime control forecasts.
            rt_pf_kwargs (dict): Additional request keywords passed to realtime perturbed forecasts.
            hc_cf_kwargs (dict): Additional request keywords passed to hindcast control forecasts.
            hc_pf_kwargs (dict): Additional request keywords passed to hindcast perturbed forecasts.
            **kwargs (): Additional request keywords.
        """
        self.request["param"] = utils.list_to_string(list(param))
        self.request["levelist"] = utils.list_to_string(list(plevs))
        if area is not None:
            self.request["area"] = area

        # all forecast steps
        if step == "all":
            step = self.STEP_ALL
        else:
            step = utils.list_to_string(step)
        self.request["step"] = step

        # drop dates without model initialization
        filtered_reftime = self.filter_reftimes(reftime)
        if exact_reftime:
            assert filtered_reftime == reftime, "reftime contains dates that are no reftimes. only allowed if " \
                                                "assert_reftime=False. "

        self.request.update(dict(**kwargs))

        for d in filtered_reftime:

            self.request["date"] = str(d)

            for fc_type, fc_type_kwargs, fc_type_class in [("rt_cf", rt_cf_kwargs, RtCf), ("rt_pf", rt_pf_kwargs, RtPf),
                                                           ("hc_cf", hc_cf_kwargs, HcCf),
                                                           ("hc_pf", hc_pf_kwargs, HcPf)]:
                if fc_type_kwargs.get("skip", False):
                    continue
                fc_type_kwargs = dict(fc_type_class(d).request, **fc_type_kwargs)
                self.request["target"] = path + "/" + self.file_name(
                    file_descr=file_descr, fc_type=fc_type, reftime=d
                )
                full_request = dict(self.request, **fc_type_kwargs)
                super().retrieve(full_request)
                if write_info_file:
                    self.make_request_info_file(path, file_descr, fc_type, full_request, filtered_reftime)
            write_info_file = False

    def file_name(self, file_descr, fc_type, reftime):
        """
        File name convention.
        Args:
            file_descr (str): Description of fields that is included in the file name.
            fc_type (str): Usually one of "rt_cf", "rt_pf", "hc_cf", "hc_pf".
            reftime (np.datetime64): Model realtime initilization date (reference time).

        Returns : File Name (str)

        """
        target = "s2s_{model}_{file_descr}_{reftime}_{fc_type}.nc".format(
            model=self.MODEL_NAME,
            file_descr=file_descr,
            fc_type=fc_type,
            reftime=reftime
        )
        return target

    def make_request_info_file(self, path, file_descr, fc_type, full_request, all_dates):
        """
        Create a .json file including information about s2stools request.
        Args:
            path (str): Directory to place file.
            file_descr (str): Description of fields that is included in file name.
            fc_type (str): Usually one of "rt_cf", "rt_pf", "hc_cf", "hc_pf".
            full_request (dict): Full API request.
            all_dates (list of np.datetime64[D]):  List of all requested reftime dates.
        """
        now = datetime.now().isoformat(timespec="minutes")
        filename = "request_s2s_{model_name}_{file_descr}_{datetime}_{fc_type}.json".format(
            model_name=self.MODEL_NAME,
            file_descr=file_descr,
            datetime=now.replace(":", ""),
            fc_type=fc_type
        )
        content = {
            "info": {
                "time": now,
                "fc_type": fc_type,
                "reftime_dates": list(all_dates.astype("str"))
            },
            "request": full_request
        }

        if True:  # for development purpose
            with open(path + "/" + filename, "w") as outfile:
                json.dump(content, outfile)

    @staticmethod
    def filter_reftimes(dates):
        """
        Drop dates that are no valid reftimes (realtime model initialization dates) for S2S ECMWF.
        Args:
            dates (list of np.datetime64[D]):  Dates to analyze.
        Returns:
            list: List of valid dates.
        """
        dates = np.atleast_1d(dates)
        dates_pd = pd.Series(dates)
        filtered_dates = dates[(dates_pd.dt.weekday == 0) | (dates_pd.dt.weekday == 3)]
        assert len(filtered_dates) > 0, "no dates left after filtering reftimes"
        return filtered_dates



import numpy as np
from datetime import datetime
from .. import utils


def date_to_winter_season(date):
    if date.month <= 6:
        y1 = date.year - 1
        y2 = date.year
    else:
        y1 = date.year
        y2 = date.year + 1
    return str(y1)[:] + "/ " + str(y2)[-2:]


def annotate_bars(ax, labels):
    # labels on top of bars
    rects = ax.patches

    ylim = ax.get_ylim()
    for rect, label in zip(rects, labels):
        height = rect.get_height()
        ax.text(
            rect.get_x() + rect.get_width() / 2,
            height + np.diff(ylim) / 10,
            label,
            ha="center",
            va="bottom",
            rotation=90,
        )
    ax.set_ylim(None, ylim[1] * 1.5)


def replace_year(dt64, year):
    # convert to timestamp:
    ts = (dt64 - np.datetime64("1970-01-01T00:00:00Z")) / np.timedelta64(1, "s")
    # standard utctime from timestamp
    dt = datetime.utcfromtimestamp(ts)
    # update year
    dt = dt.replace(year=year)
    # convert back to numpy.datetime64:
    return np.datetime64(dt).astype("datetime64[D]")


def fc_dates(reftime=None, hc_year=None, leadtime=None, data=None):
    if data is not None:
        reftime, hc_year, leadtime = (
            data.reftime.values,
            data.hc_year.values,
            data.leadtime.values,
        )
    else:
        reftime = np.array(reftime)
        hc_year = np.array(hc_year)
        leadtime = np.array(leadtime)
    dates = utils.add_years(reftime, hc_year) + leadtime
    return dates



import xarray as xr


def data_statistics(data_comp):
    """compute mean, std, min, max all at once.

    Args:
        data_comp (xr.DataArray): Compsite DataArray for which statistics shall be computed, expected dimensions are "i" and "days_since_event".

    Returns:
        tuple: (mean, std, min, max), all of type xr.DataArray with dimenesion only "days_since_event".
    """
    data_comp_mean = data_comp.mean("i")
    data_comp_std = data_comp.std("i")
    data_comp_min = data_comp.min("i")
    data_comp_max = data_comp.max("i")
    return data_comp_mean, data_comp_std, data_comp_min, data_comp_max


def data_percentiles(data_comp):
    """compute percentiles 10, 30, 70, 90, 98

    Args:
        data_comp (xr.DataArray): Composite DataArray to compute percentiles, expected dimensions are "i" and "days_since_event".

    Returns:
        xr.DataArray: Composite DataArray with dimesions "quantile" and "days_since_event".
    """
    data_comp.load()

    quantiles = [2, 10, 30, 70, 90, 98]
    data_comp_quantiles = {}
    for q in quantiles:
        data_comp_quantiles[str(q)] = data_comp.quantile(q / 100, dim="i")
    return data_comp_quantiles




from unittest import TestCase
from s2stools import process, utils
import os
import xarray as xr


class TestProcess(TestCase):

    def test_open_files(self):
        directory = "../../data/"
        ds = process.open_files(path_pattern=directory + "s2s_ecmwf_uv_20*", max_lt=1)
        self.assertIsInstance(ds, xr.Dataset)

    def test_add_validtime(self):
        ds = process.open_files(path_pattern="../../data/s2s_ecmwf_uv_20*")
        print(ds)
        ds_vt = process.add_validtime(ds)
        self.assertIn("validtime", list(ds_vt.coords))
        vt_min = ds_vt.validtime.min()
        hc_year_min = ds.hc_year.min().values
        reftime_min = ds.reftime.min().values
        earliest_date = utils.add_years(reftime_min, hc_year_min)
        self.assertEqual(earliest_date, vt_min)



from .. import utils


def add_validtime(da):
    """
    number = number
    fc_day = reftime + hc_year + leadtime
    hc_year = hc_year
    lead = leadtime
    """
    da_stacked = da.stack(day=("reftime", "hc_year", "leadtime"))
    fc_day = (
            utils.add_years(da_stacked.reftime.values, da_stacked.hc_year.values)
            + da_stacked.leadtime.values
    )

    fc_day_reshaped = fc_day.reshape(
        len(da.reftime), len(da.hc_year), len(da.leadtime)
    )
    res = da.assign_coords(
        validtime=(("reftime", "hc_year", "leadtime"), fc_day_reshaped)
    )
    return res



from unittest import TestCase
from s2stools import clim, process


class TestClim(TestCase):
    
    def test_deseasonalize(self):
        data = process.open_files(path_pattern="../../data/s2s_*")
        # absolute anomalies
        anom = clim.deseasonalize(data, standardize=False)
        for var in list(anom.keys()):
            self.assertAlmostEqual(anom[var].mean().values, 0, delta=0.1)

        # standardized absolute anomalies
        anom = clim.deseasonalize(data, standardize=True)
        for var in list(anom.keys()):
            self.assertAlmostEqual(anom[var].mean().values, 0, delta=0.1)
            self.assertAlmostEqual(anom[var].std().values, 1, delta=0.1)






import xarray as xr
from tqdm import tqdm
import numpy as np


def extreme_predictors(data_comp):
    """Compute the fraction of events with nam1000 below 0, -3 and above +3.

    Args:
        data_comp (xr.Dataset): Dataset that contains variable nam1000 and dimensions "days_since_event" and "i".

    Returns:
        xr.Dataset: Dataset with dimension "days_since_event"
    """

    predictor_fractions_m2 = []
    predictor_fractions_m0 = []
    predictor_fractions_p2 = []

    for d in data_comp.days_since_event:
        data_sel = data_comp.nam1000.sel(days_since_event=d).compute()

        data_sel_extr = data_sel.where(data_sel < -3, drop=True)
        fraction = len(data_sel_extr.i) / len(data_sel.dropna("i").i)
        predictor_fractions_m2.append(fraction)

        data_sel_extr = data_sel.where(data_sel < 0, drop=True)
        fraction = len(data_sel_extr.i) / len(data_sel.dropna("i").i)
        predictor_fractions_m0.append(fraction)

        data_sel_extr = data_sel.where(data_sel > 3, drop=True)
        fraction = len(data_sel_extr.i) / len(data_sel.dropna("i").i)
        predictor_fractions_p2.append(fraction)

    return xr.Dataset(
        data_vars={
            "extremors_nam1000_m2": ("days_since_event", predictor_fractions_m2),
            "extremors_nam1000_m0": ("days_since_event", predictor_fractions_m0),
            "extremors_nam1000_p2": ("days_since_event", predictor_fractions_p2),
        },
        coords={"days_since_event": ("days_since_event", data_comp.days_since_event.values)},
    )


def prob_oneday_extreme_nam_within_period_clim(
        nam_data,
        extreme_threshold,
        bootstrap_sample_size,
        n_bootstrap_samples=1_000,
        uncertainty_percentiles=[0.025, 0.975],
        max_days=None,
):
    """Compute climatological probability of at least 1 day below threshold as function of waiting period.

    Args:
        nam_data (DataArray): data of S2S model (not stacked)
        extreme_threshold (float): threshold
        bootstrap_sample_size (int): how many forecasts to sample for one estimate
        n_bootstrap_samples (int, optional): how many times to repeat drwaing estimates. Defaults to 1_000.
        uncertainty_percentiles (list, optional): Confidence Interval. Defaults to [0.025, 0.975].
        max_days (int, optional): How many days for climatology computation, if None than =len(days_since_init). Defaults to None.

    Returns:
        DataArray: coords are {"stat": ["mean", "uncertainty_top", "uncertainty_bottom"], "days": <day_list>}
    """

    dsi_start = 10  # parameter that can be set
    if not max_days:
        max_days = len(nam_data.days_since_init)  # 35

    # iterate over period lengths

    mean = []
    uncertainty_top = []
    uncertainty_bottom = []
    for pl in tqdm(range(max_days), desc="p clim iterating period lengths"):
        # compute climatological mean and CI for p between 1 and i
        pop_mean, pop_ci = bootstrap_extr_prob(
            population=nam_data.stack(fc=("reftime", "hc_year", "number")).dropna(
                "fc", how="all"
            ),
            sample_size=bootstrap_sample_size,  # len(ssw.comp.i),
            dsi_slice=slice(dsi_start, dsi_start + pl),
            threshold=extreme_threshold,
            n_bootstrap_samples=n_bootstrap_samples,
        )
        mean.append(pop_mean)
        uncertainty_bottom.append(pop_ci[0])
        uncertainty_top.append(pop_ci[1])

    pos_lag_days = np.arange(1, max_days + 1)
    mean = np.array(mean)
    uncertainty_bottom = np.array(uncertainty_bottom)
    uncertainty_top = np.array(uncertainty_top)

    res = xr.DataArray(
        np.array([mean, uncertainty_top, uncertainty_bottom]),
        dims=["stat", "days"],
        coords={
            "stat": np.array(["mean", "uncertainty_top", "uncertainty_bottom"]),
            "days": pos_lag_days,
        },
    )
    return res


def prob_oneday_extreme_nam_within_period_after_event(nam_comp, extreme_threshold):
    """Compute probability for at least 1 day nam < extreme_threshold within period.

    Args:
        nam_comp (DataArray): comp with dims=['days_since_event', 'i']
        extreme_threshold (float): threshold

    Returns:
        DataArray: probability with variable 'days' which is the period length
    """

    period_length = len(nam_comp.sel(days_since_event=slice(1, None)).days_since_event)

    p = []
    for i in tqdm(range(period_length), desc="p clim iterating period lengths"):
        p_within_i_days_after_event = integrated_extr_prob(
            data=nam_comp, dse_slice=slice(1, 1 + i), threshold=extreme_threshold
        )
        p.append(p_within_i_days_after_event)

    pos_lag_days = np.arange(1, period_length + 1)
    p = np.array(p)
    res = xr.DataArray(p, dims=["days"], coords={"days": pos_lag_days})
    return res


def bootstrap_extr_prob(
        population, sample_size, dsi_slice, threshold, alpha=0.05, n_bootstrap_samples=1_000
):
    """Bootstrapping forecasts to compute mean and confidence interval for probability that NAM is below threshold.

    Args:
        population (DataArray): dimensions 'fc' and 'days_since_init'
        sample_size (int): size of one sample that is bootstrapped
        dsi_slice (int): at which 'days_since_init' to start to bootstrap
        threshold (int): NAM below which threshold?
        alpha (float, optional): alpha-level for confidence interval. Defaults to 0.05.
        n_bootstrap_samples (int, optional): times to repeat drawing samples from the population. Defaults to 1_000.

    Returns:
        (float, (float, float)): (mean, (CI_bottom, CI_top))
    """

    population_dsi_sliced = population.sel(days_since_init=dsi_slice)
    population_dsi_sliced_vals = population_dsi_sliced.values
    rand_fc_idx = np.random.randint(
        low=0, high=len(population.fc), size=sample_size * n_bootstrap_samples
    )
    n_dsi = len(population_dsi_sliced.days_since_init)
    population_dsi_sliced_vals_reshaped = population_dsi_sliced_vals[
                                          :, rand_fc_idx
                                          ].reshape(n_dsi, sample_size, n_bootstrap_samples)

    if threshold <= 0:
        is_extreme = (population_dsi_sliced_vals_reshaped < threshold).max(
            axis=0
        )  # is extreme on at least one day
    else:
        is_extreme = (population_dsi_sliced_vals_reshaped > threshold).max(
            axis=0
        )  # is extreme on at least one day
    n_extremes = is_extreme.sum(axis=0)  # sum over forecasts
    p_extremes = n_extremes / sample_size  # divide by sample size
    p_extremes_list = p_extremes.tolist()  # list to determine mean/ CI/...

    def empirical_cdf(array):
        N = len(array)
        x = np.sort(array)
        cdf = np.arange(N) / N
        return x, cdf

    x, cdf = empirical_cdf(p_extremes_list)
    test = np.argwhere((cdf > alpha / 2) & (cdf < 1 - alpha / 2))
    idx_low, idx_high = np.argwhere((cdf > alpha / 2) & (cdf < 1 - alpha / 2))[[0, -1]]
    ci = float(x[idx_low]), float(x[idx_high])
    mean = np.mean(p_extremes_list)
    return mean, ci


# used for p after event
def integrated_extr_prob(data, dse_slice, threshold):
    """Probability that data < threshold between dse_slice days.

    Args:
        data (DataArray): composite with coordinates i, days_since_event
        dse_slice (slice): within which time frame to look for extreme
        threshold (float): definition of extreme

    Returns:
        float: probability
    """
    # data_stacked_ssw = ssw.comp.rename(i="fc").nam1000
    data = data.rename(i="fc")

    n = []
    fc_to_exclude = []
    for dse in data.sel(days_since_event=dse_slice).days_since_event:
        if threshold <= 0:
            fc_extr = (
                data.sel(days_since_event=dse)
                    .where(data.sel(days_since_event=dse) < threshold, drop=True)
                    .fc
            )
        else:
            fc_extr = (
                data.sel(days_since_event=dse)
                    .where(data.sel(days_since_event=dse) > threshold, drop=True)
                    .fc
            )
        n_fc_extr = len(fc_extr)
        n_fc_to_exclude = len(fc_to_exclude)
        new_fc_extr = list(set(fc_extr.values) - set(fc_to_exclude))
        normalization = len(
            data.sel(
                days_since_event=dse,
                fc=np.isin(
                    data.sel(days_since_event=dse).fc, fc_to_exclude, invert=True
                ),
            )
                .dropna("fc")
                .fc
        )
        p = 0 if normalization == 0 else len(new_fc_extr) / normalization
        n.append(p)
        fc_to_exclude.extend(fc_extr.values)
        fc_to_exclude = list(set(fc_to_exclude))  # new

    return 1 - np.prod(1 - np.array(n))



import numpy as np
from .. import utils
from tqdm import tqdm
from ._process_events import *


def find_ssw(u60_10hPa, buffer_start=10, buffer_end=10, require_westwind_start=10):
    buffer_start, buffer_end, require_westwind_start = map(utils.to_timedelta64,
                                                           [buffer_start, buffer_end, require_westwind_start])

    var = u60_10hPa.squeeze()
    var_stacked = var.stack(fc=["reftime", "hc_year", "number"])
    var_stacked_valid = var_stacked.dropna("fc", how="all")
    print("number of:")
    print("\t all forecast combinations: ", len(var_stacked.fc))
    print("\t valid forecast combinations: ", len(var_stacked_valid.fc))

    fc_startwest = (
        var_stacked.where(
            (var_stacked.sel(leadtime=slice("0D", require_westwind_start - 1)) > 0)
        )
            .dropna("fc", how="any")
            .fc
    )

    print("\t forecasts start start with 10 days westwind: ", len(fc_startwest))

    events = []
    for fc in tqdm(fc_startwest[:], desc="Scanning Forecasts for Events"):
        fc_run = var_stacked.sel(fc=fc)

        # find zero crossing
        last_day = fc_run.leadtime.values[-1]
        eastwinddays = fc_run.where(
            fc_run.sel(leadtime=slice(buffer_start, last_day - buffer_end)) < 0,
            drop=True,
        ).leadtime

        if len(eastwinddays) > 0:
            events.append([list(np.atleast_1d(fc)[0]), eastwinddays.values[0]])

    event_dict = eventlist_to_dict(events)

    return event_dict


def find_dynssw(
        u_10hPa_60, buffer_start=10, buffer_end=10, require_westwind_start=10
):
    var_stacked = u_10hPa_60.stack(fc=["reftime", "hc_year", "number"])
    var_stacked_valid = var_stacked.dropna("fc", how="all")
    print("number of:")
    print("\t all forecast combinations: ", len(var_stacked.fc))
    print("\t valid forecast combinations: ", len(var_stacked_valid.fc))

    fc_startwest = (
        var_stacked.where(
            (var_stacked.sel(leadtime=slice(0, require_westwind_start - 1)) > 0)
        )
            .dropna("fc", how="any")
            .fc
    )

    print("\t forecasts start start with 10 days westwind: ", len(fc_startwest))

    events = []
    for fc in tqdm(fc_startwest[:], desc="Scanning Forecasts for Events"):
        fc_run = var_stacked.sel(fc=fc)

        # find zero crossing
        last_day = fc_run.leadtime.values[-1]
        eastwinddays = fc_run.where(
            fc_run.sel(leadtime=slice(buffer_start, last_day - buffer_end)) < 0,
            drop=True,
        ).leadtime

        if len(eastwinddays) > 0:
            TIMEDELTA = 5
            neg_lag = fc_run.sel(leadtime=(eastwinddays[0] - TIMEDELTA))
            pos_lag = fc_run.sel(leadtime=(eastwinddays[0] + TIMEDELTA))

            if neg_lag - pos_lag > 20:
                events.append([list(np.atleast_1d(fc)[0]), eastwinddays.values[0]])

    event_dict = eventlist_to_dict(events)

    return event_dict


def find_nam1000(nam1000, extr_type="", **kwargs):
    var = nam1000  # .squeeze()
    var_stacked = var.stack(fc=["reftime", "hc_year", "number"])
    var_stacked_valid = var_stacked.dropna("fc", how="all")
    print("number of:")
    print("\t all forecast combinations: ", len(var_stacked.fc))
    print("\t valid forecast combinations: ", len(var_stacked_valid.fc))

    events = []
    for fc in tqdm(var_stacked_valid.fc, desc="Scanning Forecasts for Events"):
        fc_run = var_stacked.sel(fc=fc)

        if extr_type == "negative":
            res = s2seventwriter_utils.nam_minus_phase(fc_run, **kwargs)
        elif extr_type == "positive":
            res = s2seventwriter_utils.nam_plus_phase(fc_run, **kwargs)
        else:
            raise Exception("Invalid extr_type. Must be 'negative' or 'positive'.")

        if res:
            events.append([list(np.atleast_1d(res.fc)[0]), int(res.leadtime)])

    event_dict = eventlist_to_dict(events)

    return event_dict



import pandas as pd
from ._utils import *


def n_events_by_lagtime(comp):
    """Counts the number of events contributing to the composite as a function of lag time.

    Args:
        comp (xr.Dataset): Composite Dataset

    Returns:
        list: n events as function of lagtime (length is equal to len(comp.lagtime))
    """
    res = []
    for lag in comp.lagtime:
        n = len(comp.sel(lagtime=lag).dropna("i").i)
        res.append(n)
    return res


def event_dates_from_ds(dataset, model="ecmwf"):
    event_dates = []

    for i in dataset.i[:]:
        ds_sel = dataset.sel(i=i)
        lt_nonan = ds_sel.dropna("lagtime", how="all").lagtime
        if len(lt_nonan) > 0:
            lt = lt_nonan[0]
        else:
            continue
        if model == "ecmwf":
            model_aware_reftime = ds_sel["reftime"]
        elif model == "ukmo":
            model_aware_reftime = (
                ds_sel["reftime"]
                if ds_sel["hc_year"] == 0
                else replace_year(ds_sel["reftime"], 2016)
            )
        else:
            raise Exception("Invalid model: {}. Must be 'ecmwf' or 'ukmo'.".format(model))
        date = fc_dates(
            reftime=model_aware_reftime,
            hc_year=ds_sel["hc_year"],
            leadtime=lt,
        )
        event_dates.append(date)

    return pd.Series(event_dates, dtype="datetime64[D]")


def fc_day_dates_permonth(data):
    data_stacked_days = data.stack(day=("reftime", "hc_year", "leadtime", "number")).dropna(
        "day", how="all"
    )

    fc_day_dates = (
            data_stacked_days.reftime.values.astype("datetime64[D]")
            + data_stacked_days.leadtime.values
    )
    fc_day_dates = pd.Series(fc_day_dates)
    return fc_day_dates.groupby(fc_day_dates.map(lambda t: (t.month))).count()


def event_leadtimes_from_ds(dataset):
    event_dates = []

    for i in dataset.i[:]:
        ds_sel = dataset.sel(i=i)
        lt_nonan = ds_sel.dropna("lagtime").lagtime.values
        if len(lt_nonan) > 0:
            lt = lt_nonan[0]
        else:
            continue
        event_dates.append(np.abs(lt))

    return pd.Series(event_dates, dtype="timedelta64[D]")



import matplotlib.pyplot as plt


def fill_between(dataarray, x, y, ax=None, ci=None, **kwargs):
    """
    wraps matplotlib's fill_between for DataArrays
    Parameters
    ----------
    dataarray : xr.DataArray
    x : str
        coordinate name
    y : str
        coordinate name
    ci : (tuple)
        quantiles
    ax : plt.axis
        axis to draw plot, if None call plt.fill_between()
    kwargs : dict
        kwargs passed to matplotlib's fill_between
    Returns
    -------
    void
    """
    x_vals = dataarray[x]
    if ci is None:
        y1_vals = dataarray.min(y)
        y2_vals = dataarray.max(y)
        y_q = [y1_vals, y2_vals]
    else:
        assert len(ci) == 2, "ci must be a tuple consisting of two floats"
        y_q = dataarray.quantile(q=ci, dim=y)
    if ax is None:
        plt.fill_between(x_vals, *y_q, **kwargs)
    else:
        ax.fill_between(x_vals, *y_q, **kwargs)



import pandas as pd

from ._infer_statistics import *
from ._infer_metadata import *
from ._utils import *
from ._manage import *
from ._extreme_prob import *
import matplotlib.pyplot as plt
from matplotlib import ticker
import calendar


class EventComposite:
    # attributes
    data = None
    comp = None
    descr = ""
    event_list_json = []
    plot_colors = {"primary": "darkslategray"}

    # init
    def __init__(self, data, event_jsons_path, descr, model, plot_colors=None):
        if plot_colors:
            self.plot_colors = self.plot_colors | plot_colors
        self.data = data
        self.descr = descr
        self.model = model
        # events from json
        if isinstance(event_jsons_path, str):
            # assert that event_jsons_path describes a path
            self.event_list_json = eventlist_from_json(
                event_jsons_path.replace("MODEL", model)
            )
        elif isinstance(event_jsons_path, list):
            # assert event_jsons_path is the list of events
            self.event_list_json = event_jsons_path
        else:
            raise TypeError
        # create composite from event list
        self.comp = composite_from_eventlist(self.event_list_json, data)

    def __len__(self):
        return len(self.comp.i)

    # plot functions

    def plot_composite(self, title_event_name="S2S events", save=False, save_descr=""):
        data_comp = self.comp
        n_contr_events = n_events_by_lagtime(self.comp)

        data_comp_mean = data_comp.mean("i")
        data_comp_std = data_comp.std("i")
        n_events = len(data_comp.i)

        # plot mean wind and mean nam on a relative time axis w.r.t. ssw central date
        fig, ax = plt.subplots(3, 1, figsize=(4, 5), sharex=True)
        c = self.plot_colors.get("primary")

        data_comp_mean.u60.plot(ax=ax[0], color=c)
        ax[0].fill_between(
            data_comp.lagtime.values,
            data_comp_mean.u60 + data_comp_std.u60,
            data_comp_mean.u60 - data_comp_std.u60,
            alpha=0.2,
            color=c,
        )

        ax0_2 = ax[0].twinx()
        data_comp_mean.u60anom.plot(ax=ax[0], color=c, linestyle="--", alpha=0.5)
        ax[0].set_ylim(-30, 65)
        ax[0].set_yticks(np.arange(-25, 51, 25), minor=False)
        ax[0].set_ylabel("u60, 10hPa [m/s]\n(dashed: anomaly)")
        ax[0].set_title("")
        ax0_2.plot(
            self.comp.lagtime.values,
            np.array(n_contr_events) / n_events,
            linestyle="dotted",
            c="dimgray",
            zorder=-10,
            alpha=0.6,
        )
        ax0_2.set_ylabel("events (dotted)", color="dimgray")
        ax0_2.set_yticks([0, 0.5, 1])
        ax0_2.set_yticklabels(["0%", "50%", "100%"], color="dimgray")
        ax0_2.tick_params(axis="y", colors="dimgray")

        data_comp_mean.nam200.plot(ax=ax[1], color=c)
        ax[1].fill_between(
            data_comp.lagtime.values,
            data_comp_mean.nam200 + data_comp_std.nam200,
            data_comp_mean.nam200 - data_comp_std.nam200,
            alpha=0.2,
            color=c,
        )
        ax[1].set_ylim(-2.1, 2.1)
        ax[1].set_yticks(np.arange(-2, 2.1, 0.5), minor=True)

        data_comp_mean.nam1000.plot(ax=ax[2], color=c)
        ax[2].fill_between(
            data_comp.lagtime.values,
            data_comp_mean.nam1000 + data_comp_std.nam1000 * 0.1,
            data_comp_mean.nam1000 - data_comp_std.nam1000 * 0.1,
            alpha=0.2,
            color=c,
        )
        ax[2].set_ylim(-2.1, 2.1)
        ax[2].set_yticks(np.arange(-2, 2.1, 0.5), minor=True)

        ax[0].set_xlabel("")
        ax[0].set_title("")
        ax[1].set_xlabel("")
        ax[1].set_title("")
        ax[1].set_ylabel("NAM200")
        ax[2].set_xlabel("lag [days]")
        ax[2].set_title("")
        ax[2].set_ylabel("NAM1000")

        for x in np.append(ax, ax0_2):
            x.spines["top"].set_visible(False)
            x.spines["right"].set_visible(False)
            x.spines["bottom"].set_visible(False)
            if x != ax0_2:
                x.grid(alpha=0.45)
                x.grid(which="minor", alpha=0.15)
                x.axhline(0, color="black", alpha=0.3, linewidth=1)
                x.axvline(0, color="black", alpha=0.3, linewidth=1)
                x.xaxis.set_major_locator(ticker.MultipleLocator(7))
                x.set_xlim(
                    data_comp.lagtime[0], data_comp.lagtime[-1]
                )

        fig.suptitle("Composite over {} {}".format(n_events, title_event_name))
        if save:
            fig.savefig(
                "ssw_composite4/composite_{}_{}.pdf".format(self.descr, save_descr),
                bbox_inches="tight",
            )
        return fig, np.append(ax, ax0_2)

    def plot_percentiles(self, save=False, save_kw=""):
        statistics = data_statistics(self.comp)
        quantiles = data_percentiles(self.comp)

        data_comp_mean, data_comp_std, data_comp_min, data_comp_max = statistics

        fig, ax = plt.subplots(figsize=(7, 4))

        # mean, min, max
        c_mean = "black"
        data_comp_mean.nam1000.plot(ax=ax, color="whitesmoke", lw=2, label="Mean")
        data_comp_min.nam1000.plot(ax=ax, color="black", lw=0.5, label="Min")
        data_comp_max.nam1000.plot(ax=ax, color="black", lw=0.5, label="Max")

        # quantiles
        ax.fill_between(
            self.comp.lagtime,
            data_comp_min.nam1000,
            quantiles["2"].nam1000,
            alpha=0.2,
            color="lightgrey",
            label="0-2 Percentile",
        )
        ax.fill_between(
            self.comp.lagtime,
            quantiles["10"].nam1000,
            quantiles["90"].nam1000,
            alpha=0.9,
            color="lightgrey",
            label="10-90 Percentile",
        )
        ax.fill_between(
            self.comp.lagtime,
            quantiles["30"].nam1000,
            quantiles["70"].nam1000,
            alpha=0.9,
            color="grey",
            label="30-70 Percentile",
        )
        ax.fill_between(
            self.comp.lagtime,
            quantiles["98"].nam1000,
            data_comp_max.nam1000,
            alpha=0.2,
            color="lightgrey",
            label="98-100 Percentile",
        )

        # auxilary plot lines
        ax.axhline(0, color="darkcyan", alpha=1, linewidth=1)
        ax.axvline(0, color="darkcyan", alpha=1, linewidth=1)

        # plot appearance
        ylim = max(np.abs(ax.get_ylim()))
        ax.set_ylim(-ylim, ylim)
        ax.grid(alpha=0.3)
        ax.set_xticks(np.arange(-21, 22, 7))
        ax.xaxis.set_major_locator(ticker.MultipleLocator(7))

        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["bottom"].set_visible(False)

        # legend
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5), frameon=False)

        # finalize
        if save:
            raise NotImplementedError
            # fig.savefig(
            #     DIR_PLOTS
            #     + script_name
            #     + "s2s_downcoup_{}{}{}_percentiles.pdf".format(setup, winter, save_kw)
            # )
        return fig, ax

    def plot_eventinfo(self, save=True, save_kw=""):

        event_dates = event_dates_from_ds(
            self.comp, model=self.model
        ).rename("all")
        event_dates_rt = event_dates_from_ds(
            self.comp.where(self.comp.hc_year == 0, drop=True), model=self.model
        ).rename("realtime")
        event_dates_hc = event_dates_from_ds(
            self.comp.where(self.comp.hc_year != 0, drop=True), model=self.model
        ).rename("hindcast")

        event_leadtimes = event_leadtimes_from_ds(self.comp)
        event_leadtimes_rt = event_leadtimes_from_ds(
            self.comp.where(self.comp.hc_year == 0, drop=True)
        )
        event_leadtimes_hc = event_leadtimes_from_ds(
            self.comp.where(self.comp.hc_year != 0, drop=True)
        )
        fc_day_dates_permonth_all = fc_day_dates_permonth(self.data).rename(
            "all"
        )
        fc_day_dates_permonth_rt = fc_day_dates_permonth(
            self.data.where(self.data.hc_year == 0, drop=True)
        ).rename("realtime")
        # fc_day_dates_permonth_hc = local_utils.fc_day_dates_permonth(
        #     self.data.where(self.data.hc_year != 0, drop=True)
        # ).rename("hindcast")

        # plot
        fig, axs = plt.subplots(
            2,
            2,
            figsize=(8, 5),  # (11.5, 3),
            gridspec_kw={"width_ratios": [3, 1]},
        )
        color = "teal"

        ax = axs[0, 0]  # *** BY WINTER SEASON ***

        count_seas_all = (
            event_dates.groupby(event_dates.map(date_to_winter_season))
                .count()
                .rename("all")
        )
        count_seas_rt = (
            event_dates_rt.groupby(
                event_dates_rt.map(date_to_winter_season)
            )
                .count()
                .rename("realtime")
        )
        count_seas_hc = (
            event_dates_hc.groupby(
                event_dates_hc.map(date_to_winter_season)
            )
                .count()
                .rename("hindcast")
        )

        count_seas_df = pd.concat(
            [count_seas_all, count_seas_rt, count_seas_hc], axis=1
        )
        hatch_kws = dict(color="yellowgreen", edgecolor=color, hatch="//")
        plane_kws = dict(color=color, edgecolor=color)
        # plot all forecasts hatched, then the hindcasts in solid, so that only the rt appear hatched
        count_seas_df["all"].rename("realtime").plot.bar(ax=ax, **hatch_kws)
        count_seas_df["hindcast"].plot.bar(ax=ax, **plane_kws)
        ax.set_xlabel("by winter")
        ax.set_ylabel("absolute frequency")
        # for legend
        xlim = ax.get_xlim()
        ax.bar([0], [0], color="k", edgecolor="k", label="all", linewidth=0)
        ax.set_xlim(xlim)

        ax = axs[0, 1]  # *** BY MONTH ***

        count_month_df = (
            pd.concat([pd.to_datetime(event_dates), pd.to_datetime(event_dates_rt)], axis=1)
                .applymap(lambda t: t.month)
                .apply(pd.Series.value_counts)
        )
        fc_day_dates_permonth_df = pd.concat(
            [fc_day_dates_permonth_all, fc_day_dates_permonth_rt], axis=1
        )
        ax.axis("off")

        n_fc = (
            count_month_df["all"]
                .reindex(list(range(7, 13)) + list(range(1, 7)))
                .dropna(how="all")
                .values
        )
        lbls = ["{:.0f}".format(n) for n in n_fc]  ## fc:

        ax = axs[1, 1]  # *** BY MONTH (NORMALIZED) ***

        count_month_df_hc_normalized = count_month_df / fc_day_dates_permonth_df
        count_month_df_hc_normalized = (
            count_month_df_hc_normalized[["all"]]
                .reindex(list(range(7, 13)) + list(range(1, 7)))
                .dropna(how="all")
        )
        p = count_month_df_hc_normalized.plot.bar(
            ax=ax, legend=False, width=0.6, color="k", edgecolor="k"
        )
        ax.set_xlabel("by month")
        ax.set_ylabel("relative frequency\nper day forecast")
        ax.set_title("labels: abs. freq.", fontdict={"size": 10})
        annotate_bars(ax, lbls)

        # by month abs and norm
        def labels_nums_to_months(labels):
            month_dict = {
                index: month for index, month in enumerate(calendar.month_abbr)
            }
            labels_num = [t.get_text() for t in labels]
            labels_str = list(map(lambda x: month_dict[int(x)], labels_num))
            return labels_str

        for ax in [axs[1, 1]]:  # [axs[0, 1], axs[1, 1]]:
            for patch in ax.patches:
                patch.set_hatch("//")
            labels_str = labels_nums_to_months(ax.get_xticklabels())
            ax.set_xticklabels(labels_str)

        ax = axs[1, 0]  # *** BY LEADTIME ***

        count_leadtime = event_leadtimes.groupby(event_leadtimes).count()
        # count_leadtime_rt = event_leadtimes_rt.groupby(event_leadtimes_rt).count()
        count_leadtime_hc = event_leadtimes_hc.groupby(event_leadtimes_hc).count()

        # count_leadtime.plot.bar(ax=ax, color=color, edgecolor=color)
        count_leadtime.plot.bar(ax=ax, color="yellowgreen", edgecolor=color, hatch="//")
        count_leadtime_hc.plot.bar(ax=ax, color=color)
        ax.set_xlabel("by leadtime [days]")
        ax.set_ylabel("absolute frequency")
        ax.xaxis.set_major_locator(ticker.MultipleLocator(5))

        for x in axs.flatten():
            x.spines["top"].set_visible(False)
            x.spines["left"].set_visible(False)
            x.spines["right"].set_visible(False)
            x.tick_params(length=0)
            x.grid(axis="y", alpha=0.5)
        fig.tight_layout()
        fig.subplots_adjust(top=0.92, wspace=0.3)

        handles, labels = axs[0, 0].get_legend_handles_labels()
        labels = [
            "realtime ({})".format(len(event_dates_rt)),
            "hindcast ({})".format(len(event_dates_hc)),
            "all ({})".format(len(event_dates)),
        ]

        axs[1, 1].legend(
            handles[::-1],
            labels[::-1],
            loc="lower center",
            bbox_to_anchor=(0.5, 1.9),
            frameon=False,
            ncol=1,
        )

        if save:
            pass
            # fig.savefig(
            #     DIR_PLOTS
            #     + script_name
            #     + "s2s_downcoup_{}{}_eventinfo.pdf".format(setup, save_kw)
            # )
            # return
            #        event_dates,
            #        event_dates_rt,
            #        count_month_df,
            #        fc_day_dates_permonth_df,
        return fig, axs

    def extreme_probability_plot(
            self,
            save=True,
            save_kw="",
            title_event_name="S2S events",
    ):

        extreme_probability = extreme_predictors(self.comp)
        dse = self.comp.lagtime.values

        def compute_base_probabilities(thresholds):
            res = []
            stacked_days = self.data.stack(
                days=("reftime", "hc_year", "number", "leadtime")
            )
            n_tot = len(stacked_days.dropna("days").days)
            for thr in thresholds:
                if thr <= 0:
                    n = len(
                        stacked_days.where(stacked_days.nam1000 < thr, drop=True).days
                    )
                else:
                    n = len(
                        stacked_days.where(stacked_days.nam1000 > thr, drop=True).days
                    )
                res.append(n / n_tot)
            return res

        base_probabilities = compute_base_probabilities(thresholds=[0, -3, 3])

        fig, ax = plt.subplots(3, 1, figsize=(5, 5), sharex=True)

        c = self.plot_colors.get("primary")  # "darkslategray"

        da_m0 = extreme_probability.extremors_nam1000_m0
        da_m2 = extreme_probability.extremors_nam1000_m2
        da_p2 = extreme_probability.extremors_nam1000_p2

        # ****************
        x = ax[0]

        plt1 = da_m0.plot(ax=x, color=c)
        if base_probabilities:
            x.axhline(base_probabilities[0], color=c, alpha=0.5, linestyle="--")
            x.fill_between(dse, da_m0, base_probabilities[0], color=c, alpha=0.2)
        x.set_ylabel("NAM1000 < 0")
        x.set_ylim(base_probabilities[0] * 0.7, base_probabilities[0] * 1.3)
        x.text(
            dse[-1] + 1,
            base_probabilities[0] * 1.01,
            "{:.3f}".format(base_probabilities[0]),
            color=c,
        )
        x.spines["bottom"].set_visible(False)

        # ****************
        x = ax[1]

        plt2 = da_m2.plot(ax=x, color=c)
        if base_probabilities:
            x.axhline(base_probabilities[1], color=c, alpha=0.5, linestyle="--")
            x.fill_between(dse, da_m2, base_probabilities[1], color=c, alpha=0.2)
        x.set_ylabel("NAM1000 < −3")
        x.set_ylim(0, 0.016)
        x.text(
            dse[-1] + 1,
            base_probabilities[1] * 1.01,
            "{:.3f}".format(base_probabilities[1]),
            color=c,
        )
        # x.annotate("", (0, 0.012), (28, 0.012), arrowprops={"arrowstyle": "<->"})
        # integrated_prob = 1 - float(
        #     (1 - da_m2.sel(lagtime=slice(1, 28))).prod("lagtime")
        # )
        # integrated_prob_clim = 1 - float((1 - base_probabilities[1]) ** 28)
        # x.text(
        #     14,
        #     0.013,
        #     "{:.1%} ↗︎ {:.1%}".format(integrated_prob_clim, integrated_prob),
        #     ha="center",
        # )

        # ****************
        x = ax[2]

        plt2 = da_p2.plot(ax=x, color=c)
        if base_probabilities:
            x.axhline(base_probabilities[2], color=c, alpha=0.5, linestyle="--")
            x.fill_between(dse, da_p2, base_probabilities[2], color=c, alpha=0.2)
        x.set_ylabel("NAM1000 > +3")
        x.set_ylim(0, 0.016)
        x.text(
            dse[-1] + 1,
            base_probabilities[2] * 1.01,
            "{:.3f}".format(base_probabilities[2]),
            color=c,
        )

        for x in ax:
            x.spines["top"].set_visible(False)
            x.spines["right"].set_visible(False)
            x.grid(alpha=0.4)
            x.set_xlabel("")
            x.xaxis.set_major_locator(ticker.MultipleLocator(7))
            x.tick_params(axis="both", which="both", length=0)
            x.axvline(0, color="black", alpha=0.3, lw=1)
        ax[2].set_xlabel("lagtime")

        fig.tight_layout()
        fig.suptitle("fraction of {} with extreme NAM1000".format(title_event_name))
        fig.subplots_adjust(top=0.94, wspace=0.4)

        if save:
            pass
            # fig.savefig(
            #     DIR_PLOTS
            #     + script_name
            #     + "s2s_downcoup_{}_extremes.pdf".format(save_kw),
            #     bbox_inches="tight",
            # )

        return fig, ax

    # in development
    def plot_prob_oneday_extreme_within_period(
            self,
            extreme_threshold,
            fig_ax=None,
            p_clim_kws={},
            p_after_event_kws={},
            lineplot_kws={},
            hide_print=True,
    ):
        """Create a plot with period length on x-Axis and Porbability of at least one day
        with NAM < threshold on y-axis. Include Climatology of Model and the Probability
        after an event.

        Args:
            extreme_threshold (float): P( nam < threshold )
            fig_ax ((plt.figure, plt.axis), optional): Existing fig and ax can be provided as tuple. Defaults to None.
            p_clim_kws (dict, optional): Keywords passed to 'prob_oneday_extreme_nam_within_period_clim'. Defaults to {}.
            p_after_event_kws (dict, optional): Keywords passed to 'prob_oneday_extreme_nam_within_period_after_event'. Defaults to {}.
            lineplot_kws (dict, optional): Keywords passed to DataArray.plot(). Defaults to {}.
            hide_print (bool, optional): Hide status prints. Defaults to True.

        Returns:
            (plt.figure, plt.axis): fig, ax of plot
        """

        ### COMPUTATION

        # compute own climatology: mean and uncertainty
        if not hide_print:
            print("Compute Climatological P...")
        p_clim_kws_updated = (
                dict(
                    nam_data=self.data.nam1000,
                    extreme_threshold=extreme_threshold,
                    bootstrap_sample_size=len(self.comp.i),
                    n_bootstrap_samples=1_000,
                )
                | p_clim_kws
        )
        p_clim = prob_oneday_extreme_nam_within_period_clim(
            **p_clim_kws_updated
        )
        # compute appearance in comp for positive lags
        if not hide_print:
            print("Compute P After Event...")
        p_after_event_kws = (
                dict(nam_comp=self.comp.nam1000, extreme_threshold=extreme_threshold)
                | p_after_event_kws
        )
        p_after_event = prob_oneday_extreme_nam_within_period_after_event(
            **p_after_event_kws
        )

        ### PLOTTING
        if not hide_print:
            print("Plot...")
        if fig_ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(5.5, 4))
        else:
            fig, ax = fig_ax

        c_prim = self.plot_colors.get("primary")
        plot_kws_clim = {"c": c_prim, "ls": "solid", "lw": 2, "ax": ax} | lineplot_kws
        plot_kws_after_event = {
                                   "c": c_prim,
                                   "ls": "--",
                                   "lw": 2,
                                   "ax": ax,
                               } | lineplot_kws

        # plot p climatology
        p_clim.sel(stat="mean").plot(
            **plot_kws_clim,
            label="S2S {}: climatology (incl. 95%-CI)".format(self.model),
        )
        x = p_clim.days.values
        y1 = p_clim.sel(stat="uncertainty_bottom").values
        y2 = p_clim.sel(stat="uncertainty_top").values
        ax.fill_between(x, y1, y2, color=plot_kws_clim.get("c"), alpha=0.15, zorder=-3)

        # plot p after event
        key = "after_event"
        p_after_event.plot(
            **plot_kws_after_event,
            label="S2S {}: following {}".format(self.model, self.descr),
        )

        # plot appearance
        ax.minorticks_on()
        ax.set_xlim(1)
        ax.set_ylim(0)
        ax.grid(alpha=0.5)
        ax.set_xlabel("t [d]")
        ax.set_ylabel("probability")
        ax.set_title(
            "$P( \geq 1 $ day $nam1000<{}$ within days 1...t $)$".format(
                extreme_threshold
            )
        )
        ax.xaxis.set_major_locator(plt.MultipleLocator(5))
        ax.tick_params(axis="both", direction="in", which="both")
        ax2 = ax.secondary_yaxis("right")
        ax2.minorticks_on()
        ax2.tick_params(axis="y", direction="in", which="both")
        ax2.set_yticklabels(ax2.get_yticklabels())

        return fig, ax



import datetime
import matplotlib.pyplot as plt
import matplotlib.ticker


def xaxis_unit_days(ax, multiple=7, minor_multiple=1):
    # Function that formats the axis labels
    def timeTicks(x, pos):
        seconds = x / 10 ** 9  # convert nanoseconds to seconds
        # create datetime object because its string representation is alright
        d = datetime.timedelta(seconds=seconds)
        return str(d.days)

    locator = plt.MultipleLocator(8.64e+13 * multiple)
    ax.xaxis.set_major_locator(locator)
    min_locator = plt.MultipleLocator(8.64e+13 * minor_multiple)
    ax.xaxis.set_minor_locator(min_locator)
    formatter = matplotlib.ticker.FuncFormatter(timeTicks)
    ax.xaxis.set_major_formatter(formatter)



from ._utils import *
from ._format import *
import matplotlib.pyplot as plt
import numpy as np


def composite_overview(ds):
    var_list = list(ds.data_vars)
    n_var = len(var_list)
    fig, axes = plt.subplots(nrows=n_var, figsize=(3, n_var * 1.3), sharex=True)
    axes = np.atleast_1d(axes)
    ds_mean = ds.mean("i")
    for i, v in enumerate(var_list):
        ax = axes[i]
        ds_mean[v].plot(ax=ax, color="k", lw=2)
        xaxis_unit_days(ax, multiple=14)
        ax.set_xlabel("")
        ax.set_title("")
        ax.axhline(0, lw=1, color="gray", zorder=-5)
        ax.axvline(0, lw=1, color="gray", zorder=-5)
        if "nam" in v:
            symmetric_ylim(ax)

    axes[0].set_title("{} events".format(len(ds.i)))
    axes[-1].set_xlabel("lagtime [days]")
    fig.tight_layout()
    return fig, ax



import numpy as np
import xarray as xr


def deseasonalize(
        data,
        window_size=15,
        standardize=False,
        ndays_clim_filter=7,
        hide_print=True,
        hide_plot=True,
        hide_warnings=False,
        return_clim_lists=False,
):
    data_by_reftime_list = []
    clim_list = []
    climstd_list = []

    for reftime in data.reftime:
        reftime = reftime.values.astype("datetime64[D]")
        window = np.arange(reftime - window_size, reftime + window_size)
        hc_reftimes_in_window = (
            data.sel(reftime=np.in1d(data.reftime.values, window))
                .sel(hc_year=-1)
                .dropna("reftime", how="all")
                .reftime.values
        )

        # ***** COMPUTE CLIMATOLOGY *****
        clim_stacked = (
            data.sel(reftime=hc_reftimes_in_window, hc_year=(data.hc_year != 0))
                .mean(["hc_year", "number"])
                .stack(date=("reftime", "leadtime"))
        )

        clim = (
            clim_stacked.assign_coords(
                days_since_reftime_init=(
                    "date",
                    (
                            clim_stacked.reftime.values - reftime + clim_stacked.leadtime.values
                    ).astype("timedelta64[D]")
                    ,
                )
            )
                .unstack()
                .groupby("days_since_reftime_init")
                .mean()
                .rolling(
                days_since_reftime_init=ndays_clim_filter, center=True, min_periods=1
            )
                .mean()
                .sel(days_since_reftime_init=slice(np.timedelta64(0, "D"), data.leadtime[-1]))
                .rename(days_since_reftime_init="leadtime")
        )
        if len(clim.leadtime) != len(data.leadtime):
            if not hide_warnings:
                print(
                    "WARNING: no climatology available for {} days of reftime {}. Applying interpolation.".format(
                        len(data.leadtime) - len(clim.leadtime), reftime
                    )
                )
            clim = clim.interp(
                leadtime=data.leadtime,
                kwargs={"fill_value": "extrapolate"},
            )

        if not hide_print:
            print(
                "climatology for reftime {}: n refs = {}".format(
                    reftime, len(hc_reftimes_in_window)
                )
            )
        if not hide_plot:
            clim.plot(label=reftime)

        if standardize:
            # ***** COMPUTE STD CLIMATOLOGY *****

            climstd_stacked = (
                data.sel(reftime=hc_reftimes_in_window, hc_year=(data.hc_year != 0))
                    .sel(number=0)
                    .stack(date=("reftime", "leadtime"))
            )

            climstd = (
                climstd_stacked.assign_coords(
                    days_since_reftime_init=(
                        "date",
                        (
                                climstd_stacked.reftime.values.astype(
                                    "datetime64[D]") - reftime + climstd_stacked.leadtime.values
                        ).astype("timedelta64[D]")
                        ,
                    )
                )
                    .unstack()
                    .groupby("days_since_reftime_init")
                    .std(["stacked_reftime_leadtime", "hc_year"])
                    .rolling(
                    days_since_reftime_init=ndays_clim_filter,
                    center=True,
                    min_periods=1,
                )
                    .mean()
                    .sel(days_since_reftime_init=slice(np.timedelta64(0, "D"), data.leadtime[-1]))
                    .rename(days_since_reftime_init="leadtime")
            )

            if len(climstd.leadtime) != len(data.leadtime):
                if not hide_warnings:
                    print(
                        "WARNING: no std climatology available for {} days of reftime {}. Applying interpolation.".format(
                            len(data.leadtime) - len(climstd.leadtime),
                            reftime,
                        )
                    )
                climstd = climstd.interp(
                    leadtime=data.leadtime,
                    kwargs={"fill_value": "extrapolate"},
                )

            data_anomstd_oneref = (data.sel(reftime=reftime) - clim) / climstd
            data_by_reftime_list.append(data_anomstd_oneref)
            climstd_list.append(climstd)

        else:
            data_anom_oneref = data.sel(reftime=reftime) - clim
            data_by_reftime_list.append(data_anom_oneref)

        clim_list.append(clim)

    anom = xr.concat(data_by_reftime_list, dim="reftime").transpose(*data.dims)

    if return_clim_lists:
        return anom, clim_list, climstd_list
    else:
        return anom


def nam(z):
    z_sel = z.sortby("latitude").sel(latitude=slice(65, 90))
    weights = np.cos(np.deg2rad(z.latitude))
    avg = (z_sel * weights).mean(["longitude", "latitude"])
    res = -1 * deseasonalize(avg, standardize=True)
    return res



from setuptools import setup

VERSION = '0.0.10'
DESCRIPTION = 'python tools for handling s2s forecast data'
LONG_DESCRIPTION = 'python tools for handling s2s forecast data.' \
                   'Including submodules: ' \
                   '<download> for retreiving s2s forecasts, ' \
                   '<clim> for deseasonalization, ' \
                   '<events> for creating event composites, ' \
                   '<plot> for matplotlib utilities, ' \
                   '<process> for handling data structures, ' \
                   '<utils> for further utilities and <tests>'

setup(
    name='s2stools',
    version=VERSION,
    author='Jonas Spaeth',
    author_email='jonas.spaeth@physik.uni-muenchen.de',
    packages=['s2stools', 's2stools.clim', "s2stools.download", "s2stools.download.ecmwf", "s2stools.events",
              "s2stools.plot", "s2stools.process", "s2stools.tests", "s2stools.utils"],
    scripts=[],  # ['bin/script1','bin/script2'],
    # url='http://pypi.python.org/pypi/PackageName/',
    license='LICENSE.txt',
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    install_requires=[
        "numpy",
        "xarray",
        "ecmwf-api-client"
    ],
)



import calendar
import numpy as np
import pandas as pd
import calendar


def list_to_string(l):
    if isinstance(l, str):
        return l
    if isinstance(l, range) | isinstance(l, np.ndarray):
        l = list(l)
    if isinstance(l, list):
        return "/".join([str(i) for i in l])
    else:
        assert False, "list_to_string excepts only type string or list, not {}".format(type(l))


def add_years(dt64, years):
    """
    Add year to date.
    Args:
        dt64 ([datetime64], datetime64): date to change
        years ([int], int): how many years to add. If scalar, add same year to each date, if list then years must have same shape as dt64.

    Returns:
        new date(s) with same shape as dt64
    """
    pd_dt = pd.to_datetime(dt64)
    if len(np.atleast_1d(years)) == 1:

        def replace_year(x):
            if (
                    (x.month == 2)
                    & (x.day == 29)
                    & (calendar.isleap(x.year + years) == False)
            ):
                return x.replace(year=x.year + years, day=28)
            else:
                return x.replace(year=x.year + years)

        res = pd.Series(pd_dt).apply(replace_year)
    else:
        assert np.array(dt64).shape == np.array(years).shape
        res = []
        for i in range(len(dt64)):
            pd_dt = pd.to_datetime(dt64[i])
            if (
                    (pd_dt.month == 2)
                    & (pd_dt.day == 29)
                    & (calendar.isleap(pd_dt.year + years[i]) == False)
            ):
                res.append(pd_dt.replace(year=pd_dt.year + years[i], day=28))
            else:
                res.append(pd_dt.replace(year=pd_dt.year + years[i]))

    res = np.array(res, "datetime64[D]")
    if len(res) == 1:
        res = res[0]
    return res


def to_timedelta64(a, assume="D"):
    """
    Assume
    Args:
        a (int or np.timedelta64): timedelta
        assume (): timedelta64 format that is assumed if a is of type int.

    Returns:
        np.timedelta64
    """
    if isinstance(a, int):
        a = np.timedelta64(a, "D")
    return a


def month_int_to_abbr(months):
    """
    convert e.g. [12, 1, 2] to ["Dec", "Jan", "Feb"]
    Args:
        months: one month or list of months as integers

    Returns:
        list of str
    """
    months = np.atleast_1d(months)
    d = {index: month for index, month in enumerate(calendar.month_abbr) if month}
    return [d[m] for m in months]



from tqdm import tqdm
import numpy as np
import pandas as pd
import xarray as xr
import json
from glob import glob
import copy


def composite_from_eventlist(event_list, data):
    event_comp = []
    missing_reftime_keys = []
    for event in tqdm(event_list, desc="Opening Events"):
        central_day = np.timedelta64(pd.Timedelta(event["leadtime"]))
        if np.datetime64(event["fc"]["reftime"]) in data.reftime.values:
            forecast = data.sel(
                reftime=event["fc"]["reftime"],
                hc_year=event["fc"]["hc_year"],
                number=event["fc"]["number"],
            )
            event_comp.append(
                forecast.assign_coords(
                    leadtime=data.leadtime - central_day
                ).rename(leadtime="lagtime")
            )
        elif event["fc"]["reftime"] in missing_reftime_keys:
            continue
        else:
            missing_reftime_keys.append(event["fc"]["reftime"])
            print(
                "KeyError Warning: reftime {} not in data".format(
                    event["fc"]["reftime"]
                )
            )

    event_comp = xr.concat(event_comp, dim="i")

    return event_comp.assign_coords(i=event_comp.i)


def composite_from_json(path, data):
    event_list = eventlist_from_json(path)
    composite = composite_from_eventlist(event_list, data)
    return composite


def eventlist_from_json(path):
    """Read a bunch of json files and generate + return an event list.

    Args:
        path (str): directory of json files, supports path globbing

    Returns:
        list: event list of format [{"fc": {"reftime": None, "hc_year": None, "number": None}, "days_since_init": None}, {...}]
    """
    event_list = []
    for f in glob(path):
        with open(f) as infile:
            event_list = event_list + json.load(infile)
    return event_list


def rename_eventlist_key(eventlist, mapping):
    """

    Args:
        mapping (dict): {"<oldkey>": "<newkey>"}

    Returns:
        updated eventlist
    """
    new_eventlist = copy.deepcopy(eventlist)
    for e in new_eventlist:
        for oldname, newname in mapping.items():
            if oldname in e:
                e[newname] = e.pop(oldname)
    return new_eventlist


def format_eventlist_dayssinceinit_to_pdtimedeltastamp(eventlist):
    new_eventlist = copy.deepcopy(eventlist)
    for e in new_eventlist:
        e["leadtime"] = e.pop("days_since_init")
        e["leadtime"] = pd.Timedelta(e["leadtime"], "D").isoformat()
    return new_eventlist
